@@ -28,14 +28,30 @@
 #include <sys/poll.h>
 #include <pthread.h>
 #include <limits.h> /* For PATH_MAX */
+#include <openssl/sha.h>
 
 #include "glusterfs-fops.h" /* generated XDR values for FOPs */
 
+#ifndef IXDR_GET_LONG
+#define IXDR_GET_LONG(buf) ((long)IXDR_GET_U_INT32(buf))
+#endif
+#ifndef IXDR_PUT_LONG
+#define IXDR_PUT_LONG(buf, v) ((long)IXDR_PUT_INT32(buf, (long)(v)))
+#endif
+#ifndef IXDR_GET_U_LONG
+#define IXDR_GET_U_LONG(buf) ((u_long)IXDR_GET_LONG(buf))
+#endif
+#ifndef IXDR_PUT_U_LONG
+#define IXDR_PUT_U_LONG(buf, v) IXDR_PUT_LONG(buf, (long)(v))
+#endif
+
 #include "list.h"
 #include "locking.h"
 #include "logging.h"
 #include "lkowner.h"
 #include "compat-uuid.h"
+#include "refcount.h"
+#include "atomic.h"
 
 #define GF_YES 1
 #define GF_NO  0
@@ -69,7 +85,11 @@
 #define FNM_EXTMATCH 0
 #endif
 
+/*gets max-offset on all architectures correctly*/
+#define GF_OFF_MAX ((1ULL << (sizeof(off_t) * 8 - 1)) - 1ULL)
+
 #define GLUSTERD_MAX_SNAP_NAME  255
+#define GLUSTERFS_SOCKET_LISTEN_BACKLOG  10
 #define ZR_MOUNTPOINT_OPT       "mountpoint"
 #define ZR_ATTR_TIMEOUT_OPT     "attribute-timeout"
 #define ZR_ENTRY_TIMEOUT_OPT    "entry-timeout"
@@ -78,11 +98,14 @@
 #define ZR_STRICT_VOLFILE_CHECK "strict-volfile-check"
 #define ZR_DUMP_FUSE            "dump-fuse"
 #define ZR_FUSE_MOUNTOPTS       "fuse-mountopts"
+#define IO_THREADS_QUEUE_SIZE_KEY "io-thread-queue-size"
 
 #define GF_XATTR_CLRLK_CMD      "glusterfs.clrlk"
 #define GF_XATTR_PATHINFO_KEY   "trusted.glusterfs.pathinfo"
 #define GF_XATTR_NODE_UUID_KEY  "trusted.glusterfs.node-uuid"
+#define GF_XATTR_LIST_NODE_UUIDS_KEY "trusted.glusterfs.list-node-uuids"
 #define GF_REBAL_FIND_LOCAL_SUBVOL "glusterfs.find-local-subvol"
+#define GF_REBAL_OLD_FIND_LOCAL_SUBVOL "glusterfs.old-find-local-subvol"
 #define GF_XATTR_VOL_ID_KEY   "trusted.glusterfs.volume-id"
 #define GF_XATTR_LOCKINFO_KEY   "trusted.glusterfs.lockinfo"
 #define GF_META_LOCK_KEY        "glusterfs.lock-migration-meta-lock"
@@ -105,6 +128,8 @@
                                          strlen (x)) == 0))
 #define XATTR_IS_NODE_UUID(x) (strncmp (x, GF_XATTR_NODE_UUID_KEY,      \
                                         strlen (GF_XATTR_NODE_UUID_KEY)) == 0)
+#define XATTR_IS_NODE_UUID_LIST(x) (strncmp (x, GF_XATTR_LIST_NODE_UUIDS_KEY, \
+                                    strlen (GF_XATTR_LIST_NODE_UUIDS_KEY)) == 0)
 #define XATTR_IS_LOCKINFO(x) (strncmp (x, GF_XATTR_LOCKINFO_KEY,        \
                                        strlen (GF_XATTR_LOCKINFO_KEY)) == 0)
 
@@ -113,8 +138,12 @@
 #define GF_XATTR_LINKINFO_KEY   "trusted.distribute.linkinfo"
 #define GFID_XATTR_KEY          "trusted.gfid"
 #define PGFID_XATTR_KEY_PREFIX  "trusted.pgfid."
+#define GFID2PATH_VIRT_XATTR_KEY  "glusterfs.gfidtopath"
+#define GFID2PATH_XATTR_KEY_PREFIX  "trusted.gfid2path."
+#define GFID2PATH_XATTR_KEY_PREFIX_LENGTH 18
 #define VIRTUAL_GFID_XATTR_KEY_STR  "glusterfs.gfid.string"
 #define VIRTUAL_GFID_XATTR_KEY      "glusterfs.gfid"
+#define GF_XATTR_MDATA_KEY  "trusted.glusterfs.mdata"
 #define UUID_CANONICAL_FORM_LEN 36
 
 #define GET_ANCESTRY_PATH_KEY "glusterfs.ancestry.path"
@@ -145,6 +174,33 @@
 #define GLUSTERFS_VERSION_XCHG_KEY     "glusterfs.version.xchg"
 
 #define GLUSTERFS_INTERNAL_FOP_KEY  "glusterfs-internal-fop"
+
+/* GlusterFS Internal FOP Indicator flags
+ * (To pass information on the context in which a paritcular
+ *  fop is performed between translators)
+ * The presence of a particular flag must be treated as an
+ * indicator of the context, however the flag is added only in
+ * a scenario where there is a need for such context across translators.
+ * So it cannot be an absolute information on context.
+ */
+#define GF_INTERNAL_CTX_KEY  "glusterfs.internal-ctx"
+
+/*
+ * Always append entries to end of the enum, do not delete entries.
+ * Currently dict_set_flag allows to set upto 256 flag, if the enum
+ * needs to grow beyond this dict_set_flag has to be changed accordingly
+ */
+enum gf_internal_fop_indicator {
+        GF_DHT_HEAL_DIR       /* Index 0 in bit array*/
+};
+
+/* Todo:
+ * Add GF_FOP_LINK_FILE         0x2ULL
+ * address GLUSTERFS_MARKER_DONT_ACCOUNT_KEY and
+ * GLUSTERFS_INTERNAL_FOP_KEY with this flag
+ */
+
+
 #define DHT_CHANGELOG_RENAME_OP_KEY   "changelog.rename-op"
 
 #define ZR_FILE_CONTENT_STR     "glusterfs.file."
@@ -153,6 +209,7 @@
 #define GLUSTERFS_WRITE_IS_APPEND "glusterfs.write-is-append"
 #define GLUSTERFS_WRITE_UPDATE_ATOMIC "glusterfs.write-update-atomic"
 #define GLUSTERFS_OPEN_FD_COUNT "glusterfs.open-fd-count"
+#define GLUSTERFS_ACTIVE_FD_COUNT "glusterfs.open-active-fd-count"
 #define GLUSTERFS_INODELK_COUNT "glusterfs.inodelk-count"
 #define GLUSTERFS_ENTRYLK_COUNT "glusterfs.entrylk-count"
 #define GLUSTERFS_POSIXLK_COUNT "glusterfs.posixlk-count"
@@ -228,12 +285,10 @@
 #define GF_REPLICATE_TRASH_DIR           ".landfill"
 
 /* GlusterFS's maximum supported Auxiliary GIDs */
-/* TODO: Keeping it to 200, so that we can fit in 2KB buffer for auth data
- * in RPC server code, if there is ever need for having more aux-gids, then
- * we have to add aux-gid in payload of actors */
 #define GF_MAX_AUX_GROUPS   65535
 
 #define GF_UUID_BUF_SIZE 50
+#define GF_UUID_BNAME_BUF_SIZE (320) /* (64 + 256) */
 
 #define GF_REBALANCE_TID_KEY     "rebalance-id"
 #define GF_REMOVE_BRICK_TID_KEY  "remove-brick-id"
@@ -263,7 +318,9 @@
 #define TIER_LINKFILE_GFID           "tier-linkfile-gfid"
 #define DHT_SKIP_OPEN_FD_UNLINK     "dont-unlink-for-open-fd"
 #define DHT_IATT_IN_XDATA_KEY       "dht-get-iatt-in-xattr"
+#define DHT_MODE_IN_XDATA_KEY       "dht-get-mode-in-xattr"
 #define GET_LINK_COUNT              "get-link-count"
+#define GF_GET_SIZE                 "get-size"
 
 /*CTR and Marker requires inode dentry link count from posix*/
 #define GF_RESPONSE_LINK_COUNT_XDATA "gf_response_link_count"
@@ -271,6 +328,9 @@
 
 #define CTR_ATTACH_TIER_LOOKUP    "ctr_attach_tier_lookup"
 
+#define CLIENT_CMD_CONNECT    "trusted.glusterfs.client-connect"
+#define CLIENT_CMD_DISCONNECT "trusted.glusterfs.client-disconnect"
+
 #define GF_LOG_LRU_BUFSIZE_DEFAULT 5
 #define GF_LOG_LRU_BUFSIZE_MIN 0
 #define GF_LOG_LRU_BUFSIZE_MAX 20
@@ -282,37 +342,91 @@
 #define GF_LOG_FLUSH_TIMEOUT_MAX 300
 #define GF_LOG_FLUSH_TIMEOUT_MIN_STR "30"
 #define GF_LOG_FLUSH_TIMEOUT_MAX_STR "300"
+#define GF_LOG_LOCALTIME_DEFAULT 0
 
 #define GF_BACKTRACE_LEN        4096
 #define GF_BACKTRACE_FRAME_COUNT 7
 
-#define GF_LK_ADVISORY 0
-#define GF_LK_MANDATORY 1
+#define GF_LK_ADVISORY 0 /* maps to GLFS_LK_ADVISORY from libgfapi*/
+#define GF_LK_MANDATORY 1 /* maps to GLFS_LK_MANDATORY from libgfapi*/
+#define GF_LOCK_MODE "glusterfs.lk.lkmode"
+
+#define GF_CHECK_XATTR_KEY_AND_GOTO(key, cmpkey, errval, lbl)   \
+        do {                                                    \
+                if (key && strcmp (key, cmpkey) == 0) {         \
+                        errval = -EINVAL;                       \
+                        goto lbl;                               \
+                }                                               \
+        } while (0)                                             \
+
+#define GF_CS_OBJECT_SIZE  "trusted.glusterfs.cs.object_size"
+
+#define GF_CS_OBJECT_UPLOAD_COMPLETE "trusted.glusterfs.csou.complete"
+#define GF_CS_OBJECT_REMOTE "trusted.glusterfs.cs.remote"
+#define GF_CS_OBJECT_DOWNLOADING "trusted.glusterfs.cs.downloading"
+#define GF_CS_OBJECT_DOWNLOADED "trusted.glusterfs.cs.downloaded"
+#define GF_CS_OBJECT_STATUS "trusted.glusterfs.cs.status"
+#define GF_CS_OBJECT_REPAIR "trusted.glusterfs.cs.repair"
+
+typedef enum {
+        GF_CS_LOCAL = 1,
+        GF_CS_REMOTE = 2,
+        GF_CS_REPAIR = 4,
+        GF_CS_DOWNLOADING = 8,
+        GF_CS_ERROR = 16,
+} gf_cs_obj_state;
+
+typedef enum {
+        GF_FOP_PRI_UNSPEC = -1,         /* Priority not specified */
+        GF_FOP_PRI_HI = 0,              /* low latency */
+        GF_FOP_PRI_NORMAL,              /* normal */
+        GF_FOP_PRI_LO,                  /* bulk */
+        GF_FOP_PRI_LEAST,               /* least */
+        GF_FOP_PRI_MAX,                 /* Highest */
+} gf_fop_pri_t;
+
+static const char * const FOP_PRI_STRINGS[] = {
+        "HIGH",
+        "NORMAL",
+        "LOW",
+        "LEAST"
+};
+
+static inline const char *fop_pri_to_string (gf_fop_pri_t pri)
+{
+        if (pri < 0)
+                return "UNSPEC";
+
+        if (pri >= GF_FOP_PRI_MAX)
+                return "INVALID";
+
+        return FOP_PRI_STRINGS[pri];
+}
 
 const char *fop_enum_to_pri_string (glusterfs_fop_t fop);
-const char *fop_enum_to_string (glusterfs_fop_t fop);
 
 #define GF_SET_IF_NOT_PRESENT 0x1 /* default behaviour */
 #define GF_SET_OVERWRITE      0x2 /* Overwrite with the buf given */
 #define GF_SET_DIR_ONLY       0x4
 #define GF_SET_EPOCH_TIME     0x8 /* used by afr dir lookup selfheal */
+#define GF_AUXILLARY_PARGFID  0xd /* RIO dummy parent gfid */
 
 /* key value which quick read uses to get small files in lookup cbk */
 #define GF_CONTENT_KEY "glusterfs.content"
 
 struct _xlator_cmdline_option {
-        struct list_head    cmd_args;
-        char               *volume;
-        char               *key;
-        char               *value;
+        struct list_head   cmd_args;
+        char              *volume;
+        char              *key;
+        char              *value;
 };
 typedef struct _xlator_cmdline_option xlator_cmdline_option_t;
 
 struct _server_cmdline {
-        struct list_head  list;
+        struct list_head   list;
         char              *volfile_server;
         char              *transport;
-        int               port;
+        int                port;
 };
 typedef struct _server_cmdline server_cmdline_t;
 
@@ -322,101 +436,123 @@ typedef struct _server_cmdline server_cmdline_t;
 
 struct _cmd_args {
         /* basic options */
-        char             *volfile_server;
-        server_cmdline_t *curr_server;
-        /* List of backup volfile servers, including original */
-        struct list_head volfile_servers;
-        char             *volfile;
-        char             *log_server;
-        gf_loglevel_t    log_level;
-        char            *log_file;
-        char            *log_ident;
-        gf_log_logger_t  logger;
-        gf_log_format_t  log_format;
-        uint32_t         log_buf_size;
-        uint32_t         log_flush_timeout;
-        int32_t          max_connect_attempts;
-        char            *print_exports;
-        char            *print_netgroups;
+        char              *volfile_server;
+        server_cmdline_t  *curr_server;
+        /* List of backup  volfile servers, including original */
+        struct list_head   volfile_servers;
+        char              *volfile;
+        char              *log_server;
+        gf_loglevel_t      log_level;
+        char              *log_file;
+        char              *log_ident;
+        gf_log_logger_t    logger;
+        gf_log_format_t    log_format;
+        uint32_t           log_buf_size;
+        uint32_t           log_flush_timeout;
+        int32_t            max_connect_attempts;
+        char              *print_exports;
+        char              *print_netgroups;
+        int                print_xlatordir;
+        int                print_statedumpdir;
+        int                print_logdir;
         /* advanced options */
-        uint32_t         volfile_server_port;
-        char            *volfile_server_transport;
-        uint32_t         log_server_port;
-        char            *pid_file;
-        char            *sock_file;
-        int              no_daemon_mode;
-        char            *run_id;
-        int              debug_mode;
-        int              read_only;
-        int              acl;
-        int              selinux;
-        int              capability;
-        int              enable_ino32;
-        int              worm;
-        int              mac_compat;
-        int              fopen_keep_cache;
-        int              gid_timeout;
-        char             gid_timeout_set;
-        int              aux_gfid_mount;
+        uint32_t           volfile_server_port;
+        char              *volfile_server_transport;
+        uint32_t           log_server_port;
+        char              *pid_file;
+        char              *sock_file;
+        int                no_daemon_mode;
+        char              *run_id;
+        int                debug_mode;
+        int                read_only;
+        int                acl;
+        int                selinux;
+        int                capability;
+        int                enable_ino32;
+        int                worm;
+        int                mac_compat;
+        int                fopen_keep_cache;
+        int                gid_timeout;
+        char               gid_timeout_set;
+        int                aux_gfid_mount;
 
         /* need a process wide timer-wheel? */
-        int              global_timer_wheel;
+        int                global_timer_wheel;
 
-        struct list_head xlator_options;  /* list of xlator_option_t */
+        /* list of xlator_option_t */
+        struct list_head   xlator_options;
 
-	/* fuse options */
-	int              fuse_direct_io_mode;
-	char             *use_readdirp;
-        int              no_root_squash;
-        int              volfile_check;
-        double           fuse_entry_timeout;
-        double           fuse_negative_timeout;
-        double           fuse_attribute_timeout;
-        char            *volume_name;
-        int              fuse_nodev;
-        int              fuse_nosuid;
-        char            *dump_fuse;
-        pid_t            client_pid;
-        int              client_pid_set;
-        unsigned         uid_map_root;
-        int              background_qlen;
-        int              congestion_threshold;
-        char             *fuse_mountopts;
-        int              mem_acct;
-        int              resolve_gids;
+        /* fuse options */
+        int                fuse_direct_io_mode;
+        char              *use_readdirp;
+        int                no_root_squash;
+        int                volfile_check;
+        double             fuse_entry_timeout;
+        double             fuse_negative_timeout;
+        double             fuse_attribute_timeout;
+        char              *volume_name;
+        int                fuse_nodev;
+        int                fuse_nosuid;
+        char              *dump_fuse;
+        pid_t              client_pid;
+        int                client_pid_set;
+        unsigned           uid_map_root;
+        int                background_qlen;
+        int                congestion_threshold;
+        char              *fuse_mountopts;
+        int                mem_acct;
+        int                resolve_gids;
 
         /* key args */
-        char            *mount_point;
-        char            *volfile_id;
+        char              *mount_point;
+        char              *volfile_id;
 
         /* required for portmap */
-        int             brick_port;
-        char           *brick_name;
-        int             brick_port2;
+        int                brick_port;
+        char              *brick_name;
+        int                brick_port2;
 
         /* Should management connections use SSL? */
-        int             secure_mgmt;
+        int                secure_mgmt;
 
         /* Linux-only OOM killer adjustment */
 #ifdef GF_LINUX_HOST_OS
-        char            *oom_score_adj;
+        char              *oom_score_adj;
 #endif
+
+        /* Run this process with valgrind? Might want to prevent calling
+         * functions that prevent valgrind from working correctly, like
+         * dlclose(). */
+        int                valgrind;
+        int                localtime_logging;
+
+        /* For the subdir mount */
+        char           *subdir_mount;
+
+        char              *process_name;
+        char              *event_history;
+        int                thin_client;
+        uint32_t           reader_thread_count;
+
+        /* FUSE writeback cache support */
+        int                kernel_writeback_cache;
+        uint32_t           attr_times_granularity;
 };
 typedef struct _cmd_args cmd_args_t;
 
 
 struct _glusterfs_graph {
-        struct list_head          list;
-        char                      graph_uuid[128];
-        struct timeval            dob;
-        void                     *first;
-        void                     *top;   /* selected by -n */
-        uint32_t                  leaf_count;
-        int                       xl_count;
-        int                       id;    /* Used in logging */
-        int                       used;  /* Should be set when fuse gets
+        struct list_head   list;
+        char               graph_uuid[128];
+        struct timeval     dob;
+        void              *first;
+        void              *top;   /* selected by -n */
+        uint32_t           leaf_count;
+        int                xl_count;
+        int                id;    /* Used in logging */
+        int                used;  /* Should be set when fuse gets
                                             first CHILD_UP */
-        uint32_t                  volfile_checksum;
+        uint32_t           volfile_checksum;
 };
 typedef struct _glusterfs_graph glusterfs_graph_t;
 
@@ -432,52 +568,70 @@ typedef enum {
 
 struct tvec_base;
 
+/* reference counting for the global (per ctx) timer-wheel */
+struct gf_ctx_tw {
+        GF_REF_DECL;
+        struct tvec_base *timer_wheel; /* global timer-wheel instance */
+};
+
 struct _glusterfs_ctx {
-        cmd_args_t          cmd_args;
-        char               *process_uuid;
-        FILE               *pidfp;
-        char                fin;
-        void               *timer;
-        void               *ib;
-        struct call_pool   *pool;
-        void               *event_pool;
-        void               *iobuf_pool;
-        void               *logbuf_pool;
-        gf_lock_t           lock;
-        size_t              page_size;
-        struct list_head    graphs; /* double linked list of graphs - one per volfile parse */
-        glusterfs_graph_t  *active; /* the latest graph in use */
-        void               *master; /* fuse, or libglusterfsclient (however, not protocol/server) */
-        void               *mgmt;   /* xlator implementing MOPs for centralized logging, volfile server */
-        void               *listener; /* listener of the commands from glusterd */
-        unsigned char       measure_latency; /* toggle switch for latency measurement */
-        pthread_t           sigwaiter;
-	char               *cmdlinestr;
-        struct mem_pool    *stub_mem_pool;
-        unsigned char       cleanup_started;
-        int                 graph_id; /* Incremented per graph, value should
-                                         indicate how many times the graph has
-                                         got changed */
-        pid_t               mnt_pid; /* pid of the mount agent */
-        int                 process_mode; /*mode in which process is runninng*/
-        struct syncenv     *env;          /* The env pointer to the synctasks */
+        cmd_args_t         cmd_args;
+        char              *process_uuid;
+        FILE              *pidfp;
+        char               fin;
+        void              *timer;
+        void              *ib;
+        struct call_pool  *pool;
+        void              *event_pool;
+        void              *iobuf_pool;
+        void              *logbuf_pool;
+        gf_lock_t          lock;
+        size_t             page_size;
 
-        struct list_head    mempool_list; /* used to keep a global list of
-                                             mempools, used to log details of
-                                             mempool in statedump */
-        char               *statedump_path;
+        /* one per volfile parse */
+        struct list_head   graphs;
 
-        struct mem_pool    *dict_pool;
-        struct mem_pool    *dict_pair_pool;
-        struct mem_pool    *dict_data_pool;
+        /* the latest graph in use */
+        glusterfs_graph_t *active;
+
+        /* fuse or nfs (but not protocol/server) */
+        void              *master;
+
+        /* xlator implementing MOPs for centralized logging, volfile server */
+        void              *mgmt;
+
+        /* listener of the commands from glusterd */
+        void              *listener;
+
+        /* toggle switch for latency measurement */
+        unsigned char      measure_latency;
+        pthread_t          sigwaiter;
+        char              *cmdlinestr;
+        struct mem_pool   *stub_mem_pool;
+        unsigned char      cleanup_started;
+        int                graph_id; /* Incremented per graph, value should
+                                        indicate how many times the graph has
+                                        got changed */
+        pid_t              mnt_pid; /* pid of the mount agent */
+        int                process_mode; /*mode in which process is runninng*/
+        struct syncenv    *env;          /* The env pointer to the synctasks */
+
+        struct list_head   mempool_list; /* used to keep a global list of
+                                            mempools, used to log details of
+                                            mempool in statedump */
+        char              *statedump_path;
+
+        struct mem_pool   *dict_pool;
+        struct mem_pool   *dict_pair_pool;
+        struct mem_pool   *dict_data_pool;
 
         glusterfsd_mgmt_event_notify_fn_t notify; /* Used for xlators to make
                                                      call to fsd-mgmt */
-        gf_log_handle_t     log; /* all logging related variables */
+        gf_log_handle_t    log; /* all logging related variables */
 
-        int                 mem_acct_enable;
+        int                mem_acct_enable;
 
-        int                 daemon_pipe[2];
+        int                daemon_pipe[2];
 
         struct clienttable *clienttable;
 
@@ -490,7 +644,12 @@ struct _glusterfs_ctx {
          * any sense, but it's not worth turning the codebase upside-down to
          * fix it.  Thus, an int.
          */
-        int                 secure_mgmt;
+        int                secure_mgmt;
+
+        /* The option is use to set cert_depth while management connection
+           use SSL
+         */
+        int                ssl_cert_depth;
 
         /*
          * Should *our* server/inbound connections use SSL?  This is only true
@@ -498,35 +657,58 @@ struct _glusterfs_ctx {
          * and SSL is set on the I/O path.  It should never be set e.g. for
          * NFS.
          */
-        mgmt_ssl_t          secure_srvr;
+        mgmt_ssl_t         secure_srvr;
         /* Buffer to 'save' backtrace even under OOM-kill like situations*/
-        char btbuf[GF_BACKTRACE_LEN];
+        char               btbuf[GF_BACKTRACE_LEN];
 
-        pthread_mutex_t notify_lock;
-        pthread_cond_t notify_cond;
-        int notifying;
+        pthread_mutex_t    notify_lock;
+        pthread_cond_t     notify_cond;
+        int                notifying;
 
-        struct tvec_base *timer_wheel; /* global timer-wheel instance */
+        struct gf_ctx_tw  *tw; /* refcounted timer_wheel */
 
+        gf_lock_t          volfile_lock;
+
+        /* configuration related elements, which gets changed
+           from global xlator */
+        struct {
+                char *metrics_dumppath;
+        } config;
+
+        struct {
+                gf_atomic_t max_dict_pairs;
+                gf_atomic_t total_pairs_used;
+                gf_atomic_t total_dicts_used;
+        } stats;
+
+        struct list_head   volfile_list;
 };
 typedef struct _glusterfs_ctx glusterfs_ctx_t;
 
+typedef struct {
+        char               volfile_checksum[SHA256_DIGEST_LENGTH];
+        char               vol_id[NAME_MAX+1];
+        struct list_head   volfile_list;
+
+} gf_volfile_t;
+
+
 glusterfs_ctx_t *glusterfs_ctx_new (void);
 
 struct gf_flock {
-        short        l_type;
-        short        l_whence;
-        off_t        l_start;
-        off_t        l_len;
-        pid_t        l_pid;
-        gf_lkowner_t l_owner;
+        short              l_type;
+        short              l_whence;
+        off_t              l_start;
+        off_t              l_len;
+        pid_t              l_pid;
+        gf_lkowner_t       l_owner;
 };
 
 typedef struct lock_migration_info {
-        struct list_head        list;
-        struct gf_flock         flock;
-        char                   *client_uid;
-        uint32_t                lk_flags;
+        struct list_head   list;
+        struct gf_flock    flock;
+        char              *client_uid;
+        uint32_t           lk_flags;
 } lock_migration_info_t;
 
 #define GF_MUST_CHECK __attribute__((warn_unused_result))
@@ -569,9 +751,12 @@ int glusterfs_graph_init (glusterfs_graph_t *graph);
 glusterfs_graph_t *glusterfs_graph_new (void);
 int glusterfs_graph_reconfigure (glusterfs_graph_t *oldgraph,
                                   glusterfs_graph_t *newgraph);
-int glusterfs_graph_attach (glusterfs_graph_t *orig_graph, char *path);
+int glusterfs_graph_attach (glusterfs_graph_t *orig_graph, char *path,
+                            glusterfs_graph_t **newgraph);
+int glusterfs_graph_parent_up (glusterfs_graph_t *graph);
 
 void
 gf_free_mig_locks (lock_migration_info_t *locks);
 
+int glusterfs_read_secure_access_file (void);
 #endif /* _GLUSTERFS_H */
